/* Specification of the echo algorithm in Alloy 6
   using graphs with a dedicated initiating node
	 and the event reification idiom
   Use with echo_reif.thm
   author: Burkhardt Renz, Nocember 2021
*/
open util/graph[Node]

let unchanged[r] { (r)' = (r) } 

enum Color {Red, Green}

abstract sig Node {
	neighbors: set Node,
	var parent: lone Node,
	var inbox: set Node,
	var color: Color
}	

one sig INode extends Node{}
sig PNode extends Node{}

fact {
	noSelfLoops[neighbors]
	undirected[neighbors]
	rootedAt[neighbors, INode]
}

pred init {
	parent = INode->INode
	no inbox
	color = Node->Red
}

pred stutter {
	unchanged[parent]
	unchanged[inbox]
	unchanged[color]
}

pred broadcast[n, fp: Node] {
	all q: n.neighbors - fp | q.inbox' = q.inbox + n
	all u: Node - n.neighbors + fp | u.inbox' = u.inbox	
}

pred initiate {
		init
		broadcast[INode, INode]
		unchanged[parent]
		unchanged[color]
}

pred forward [n: Node, msg: Node] {
	no n.parent and msg in n.inbox
	parent' = parent + n->msg
	broadcast[n, msg]
	unchanged[color]
}

pred echo [n: Node] {
	one n.parent and n.inbox = n.neighbors and n.color = Red
	unchanged[parent]
	n = INode implies
		inbox' = inbox
	else
		inbox' = inbox ++ n.parent->(n.parent.inbox + n)
		color' = color ++ (n->Green)
}

fact trans {
	init
	always {
		some transPreds
		/*
		initiate or
		some n, msg: Node | forward[n, msg] or
		some n: Node | echo[n] or
		stutter */ }
}

enum TransPred { Initiate, Forward, Echo, Stutter }

// Namensraum von Funktion und Prädikaten ist disjunkt
fun initiate: TransPred {
	{ tp: Initiate | initiate }
}
fun forward: TransPred -> Node -> Node {
	{ tp: Forward, n, msg: Node | forward[n, msg] }
}
fun echo: TransPred -> Node {
	{ tp: Echo, n: Node | echo[n] }
}
fun stutter: TransPred { 
	{ tp: Stutter | stutter}
}
fun transPreds: set TransPred {
	initiate + forward.Node.Node + echo.Node +  stutter
}

fun finished: set Node {
	{ n: Node | n.color = Green }
}

pred fairness {
	all n, msg: Node {
		(eventually always init) 
				implies (always eventually initiate)
		(eventually always (no n.parent and msg in n.inbox))
				implies (always eventually forward[n, msg])
		(eventually always (one n.parent and n.inbox = n.neighbors and n.color = Red))
				implies (always eventually echo[n])
	}
}
		
run examples {}
						
run withProgress {fairness}

assert EchoComesBack {
	fairness implies eventually INode.color = Green
}

check EchoComesBack

assert SpanningTree {
	let pt = parent - INode->INode |
		always INode.color = Green 
			implies (tree[~pt] and parent.Node = Node)
}

check SpanningTree

